/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <libintl.h>

#include "sambascanner.h"
#include "Mutex.h"
#include "Tools.h"
#include "String.h"
#include "IP_Range.h"
#include "Threads.h"
#include "Processes.h"

extern struct Options_Structure Options;

int Exec_Timeout(const char *Command, const char *Parameter, double TermTime, double KillTime, int Outfile_FD, struct Process **Processes, int *NumberProcesses) {
	pid_t PID;
	char **Arguments = NULL;
	int ArgumentCount = 0;
	double Time;
	int CurrentProcess;
					
	int Modus[2];
	struct Redirect_Thread_InputData *Data;

	message(3, _("About to execute command: %s (%s) -> FD %d\n"), Command, Parameter, Outfile_FD);

	Data = malloc(sizeof(struct Redirect_Thread_InputData));
	if (pipe(Modus) == -1) {
		message(1, _("Error creating redirection pipe for execution of '%s' (%s): %s\n"), Command, Parameter, strerror(errno));
		exit(EXIT_FAILURE);
	}
	message(4, _("Created pipe fd %d -> fd %d\n"), Modus[0], Modus[1]);
	
	Lock_Processes();
	Lock_Exec();
	Time = GetTime();
	CurrentProcess = (*NumberProcesses);
	*Processes = (struct Process*)realloc(*Processes, sizeof(struct Process) * (CurrentProcess + 1));
	((*Processes)[CurrentProcess]).Running = 1;
	((*Processes)[CurrentProcess]).Status = 0;
	((*Processes)[CurrentProcess]).Signal = 0;
	((*Processes)[CurrentProcess]).Signalling_State = 0;
	((*Processes)[CurrentProcess]).Valid = 0;
	((*Processes)[CurrentProcess]).TermTime = Time + (TermTime * Options.Timings);
	((*Processes)[CurrentProcess]).KillTime = Time + ((TermTime + KillTime) * Options.Timings);
	(*NumberProcesses)++;
	
	Data->Outfile_FD = Outfile_FD;
	Data->Read_FD = Modus[0];
	Data->ProcessID = CurrentProcess;
	Data->TermTime = TermTime;
	Data->KillTime = KillTime;
	Data->UniqueThreadID = GetUniqueThreadID();
	pthread_mutex_init(&Data->Active, NULL);
	pthread_mutex_lock(&Data->Active);
	Data->Processes = Processes;
	message(4, _("Starting thread [UniqTID %d], Redirect_Thread (ReadFD = %d -> WriteFD = %d)\n"), Data->UniqueThreadID, Data->Read_FD, Data->Outfile_FD);
	if (pthread_create_detached(&(((*Processes)[CurrentProcess]).ThreadID), Redirect_Thread, Data)) {
		message(1, _("Error creating redirection thread for execution of '%s' (%s): %s\n"), Command, Parameter, strerror(errno));
		exit(EXIT_FAILURE);
	}
	message(4, _("Thread [UniqTID = %d] received pTID = %d\n"), Data->UniqueThreadID, (int)(((*Processes)[CurrentProcess]).ThreadID));
	Unlock_Exec();
	
	PID = fork();
	if (PID == -1) {
		message(1, _("Fork failed: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (PID == 0) {
		/* Close reading end of pipe */
		close(Modus[0]);

		/* First argument is the caller */
		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = (char*)malloc((strlen(Command)*sizeof(char))+1);
		strcpy(Arguments[ArgumentCount-1], Command);
	
		/* Parse the arguments */
		{
			char *Buffer;
			int Length;
			int i = 0;
			int Escaped = 0;

			Length = strlen(Parameter);
			Buffer = (char*)malloc(Length + 1);
			if (Buffer == NULL) {
				message(1, _("Malloc failed when allocating temporary buffer: %s\n"), strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			char Tmp[2];
			Tmp[1] = 0;
			Buffer[0] = 0;
			while (i < Length) {
				if (Parameter[i] == '\'') {
					Escaped = !Escaped;
				} else if ((Parameter[i] == '\\') && (Parameter[i + 1] == '\'')) {
						Tmp[0] = '\'';
						strcat(Buffer, Tmp);	/* Add a >'< */
						i++;
				} else if ((!Escaped) && (Parameter[i] == ' ')) {
					/* Parameter was ended by space */
					ArgumentCount++;
					Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
					Arguments[ArgumentCount - 1] = (char*)malloc((strlen(Buffer) + 1) * sizeof(char));
					strcpy(Arguments[ArgumentCount - 1], Buffer);
					Buffer[0] = 0;
				} else {
					/* Either a escaped space or a regular character */
					Tmp[0] = Parameter[i];
					strcat(Buffer, Tmp);
				}
				i++;
			}
			
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
			Arguments[ArgumentCount - 1] = (char*)malloc((strlen(Buffer) + 1) * sizeof(char));
			strcpy(Arguments[ArgumentCount - 1], Buffer);

			/* Add the NULL terminator argument */
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
			Arguments[ArgumentCount - 1] = NULL;
		}

		{
			int Old_Stdout, Old_Stderr;
			int DevNull;

			DevNull = open("/dev/null", O_WRONLY, 0644);
			if (DevNull == -1) {
				message(1, _("Opening /dev/null failed: %s\n"), strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			errno = 0;
			Old_Stdout = dup(STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup1: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			Old_Stderr = dup(STDERR_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup2: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			dup2(Modus[1], STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup3: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
#ifdef DEBUG
			dup2(Modus[1], STDERR_FILENO);
#else
			dup2(DevNull, STDERR_FILENO);
#endif
			if (errno != 0) { fprintf(stderr, "Error dup4: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			execv(Command, Arguments);
			dup2(Old_Stdout, STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup5: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			dup2(Old_Stderr, STDERR_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup6: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			message(1, _("Execution of '%s' failed: %s\n"), Command, strerror(errno));
		}
		exit(EXIT_FAILURE);
	} else {
		/* Close writing end of pipe */
		close(Modus[1]);

		/* Parent process */
		((*Processes)[CurrentProcess]).PID = PID;
		((*Processes)[CurrentProcess]).Valid = 1;
		message(3, _("Forked off child '%s %s' with PID %d.\n"), Command, Parameter, PID);

	}
	message(4, _("Unlocking redirection to thread with TID %d\n"), Data->UniqueThreadID);
	pthread_mutex_unlock(&Data->Active);	/* Start redirection */
	Unlock_Processes();
	
	return CurrentProcess;
}

int WaitFor(int ProcID, struct Process **Processes) {
	double Before, After;
	int Return;
	
	Before = GetTime();
	while (((*Processes)[ProcID]).Running == 1) {
		message(5, _("Process ProcID = %d state: PID = %d, Running = %d, Return = %d, pTID = %d\n"), ProcID, (*Processes)[ProcID].PID, (*Processes)[ProcID].Running, (*Processes)[ProcID].Status, (int)((*Processes)[ProcID].ThreadID));
		usleep(100 * 1000);	/* 100ms */
	}
	After = GetTime();
	message(5, _("Process ProcID = %d left RUNNING state: PID = %d, Running = %d, Return = %d, pTID = %d\n"), ProcID, (*Processes)[ProcID].PID, (*Processes)[ProcID].Running, (*Processes)[ProcID].Status, (int)((*Processes)[ProcID].ThreadID));
	
	Return = -1;
	if (WIFSIGNALED((*Processes)[ProcID].Status)) {
		char *Msg = SignalMessage((*Processes)[ProcID].Status);
		Return = -WTERMSIG((*Processes)[ProcID].Status);
		if (Msg != NULL) {	
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, signalled by signal %s.\n"), ProcID, After - Before, Msg);
		} else {
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, signalled by signal number %d.\n"), ProcID, After - Before, -Return);
		}
	} else {
		if (WIFEXITED((*Processes)[ProcID].Status)) {
			Return = WEXITSTATUS((*Processes)[ProcID].Status);
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, exited with status %d.\n"), ProcID, After - Before, Return);
		} else {
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, exited stange (status %x).\n"), ProcID, After - Before, (*Processes)[ProcID].Status);
		}
	}

	return Return;
}

